from .cluster_test import (
    sequential_cluster_test_1samp,
    sequential_cluster_test_indep,
    sequential_cluster_test_corr
)

from .max_test import (
	sequential_permutation_t_test_1samp,
	sequential_permutation_test_indep,
	sequential_permutation_test_corr
)


from mne.utils import check_random_state
from ..cluster_test import (
    sequential_cluster_test_1samp,
    sequential_cluster_test_indep,
    sequential_cluster_test_corr
)
import numpy as np

def test_sequential_cluster_test_1samp(n = 30, seed = 0):
    rng = check_random_state(seed)
    X = rng.rand(n, 40, 40)
    look_times = np.linspace(n//3, n, 3).astype(int)
    for tail in [0, 1, -1]:
        # make sure function works
        out = sequential_cluster_test_1samp(X, look_times, n, tail = tail)
        # check output dimensions
        obs_stats, ps, adj_alpha, spending = out
        assert(adj_alpha.size == spending.size)
        assert(adj_alpha.size == ps.size)
        assert(X.shape[1:] == obs_stats[look_times[0]][0].shape)
    # test TFCE
    out = sequential_cluster_test_1samp(
        X, look_times, n,
        threshold = dict(start = 0, step = .1),
    )

def test_sequential_cluster_test_indep(n = 30, seed = 0):
    rng = check_random_state(seed)
    X = rng.rand(n, 40, 40)
    conds = rng.choice([0, 1], n)
    look_times = np.linspace(n//3, n, 3).astype(int)
    for tail in [0, 1, -1]:
        out = sequential_cluster_test_indep(X, conds, look_times, n, tail = tail)
        obs_stats, ps, adj_alpha, spending = out
        assert(adj_alpha.size == spending.size)
        assert(adj_alpha.size == ps.size)
        assert(X.shape[1:] == obs_stats[look_times[0]][0].shape)

def test_sequential_cluster_test_corr(n = 30, seed = 0):
    rng = check_random_state(seed)
    X = rng.rand(n, 40, 40)
    y = rng.rand(n)
    look_times = np.linspace(n//3, n, 3).astype(int)
    for tail in [0, 1, -1]:
        out = sequential_cluster_test_corr(X, y, look_times, n, tail = tail)
        obs_stats, ps, adj_alpha, spending = out
        assert(adj_alpha.size == spending.size)
        assert(adj_alpha.size == ps.size)
        assert(X.shape[1:] == obs_stats[look_times[0]][0].shape)



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from mne.utils.docs import docdict as docdict_mne
from typing import Callable, Dict, List, Tuple
from collections import defaultdict
import sys

##############################################################################
# Define standard documentation entries

# make defaultdict so that incorrect MNE versions don't cause fatal error in
# code even if docstrings render incorrectly
docdict = defaultdict(str, {key: docdict_mne[key] for key in docdict_mne})

docdict['look_times'] = '''
look_times : list of int
    Sample sizes at which statistical test is applied to the data, *in order*.
    Not to exceed ``max_n``.
'''

docdict['n_max'] = '''
n_max : int
    Sample size at which data collection is completed, regardless of whether the
    null hypothesis has been rejected.
'''

docdict['alpha'] = '''
alpha : float, default: 0.05
    Desired false positive rate after all looks at the data (i.e. at ``n_max``)
'''

docdict['spending_func'] = '''
spending_func : instance of SpendingFunction, default: LinearSpendingFunction
    An initialized instance of one of ``SpendingFunction``'s subclasses. This
    defines a monotonically increasing function such that f(0) = 0 and
    f(n_max) = alpha, determining how Type I error is distributed over
    sequential analyses. See [2, 3] for details and provided spending functions
    in ``niseq.spending_functions`` module.
'''

docdict['X'] = '''
X : array, shape (n_observations[, p][, q][, r])
    The data to be analyzed. The first dimension of the array is the number of
    observations; remaining dimensions comprise the size of a single observation.
    Observations must appear in the order in which they were collected.
'''

docdict['X_clust'] = '''
X : array, shape (n_observations, p[, q][, r])
    The data to be analyzed. The first dimension of the array is the number of
    observations; remaining dimensions comprise the size of a single observation.
    Observations must appear in the order in which they were collected.
    Note: that the last dimension of ``X`` should correspond to the dimension
    represented in the adjacency parameter (e.g., spectral data should be
    provided as ``(observations, frequencies, channels/vertices))``.
'''

docdict['y'] = '''
y : array, shape (n_observations,)
    Value of dependent variable associated with each observation in ``X``.
'''

docdict['tail'] = '''
tail : -1 or 0 or 1, default: 0
    If tail is 1, the alternative hypothesis is that the
    mean of the data is greater than 0 (upper tailed test).  If tail is 0,
    the alternative hypothesis is that the mean of the data is different
    than 0 (two tailed test).  If tail is -1, the alternative hypothesis
    is that the mean of the data is less than 0 (lower tailed test).
'''

docdict['tail_corr'] = '''
tail : -1 or 0 or 1, default: 0
    If tail is 1, the alternative hypothesis is that the
    correlation is greater than 0 (upper tailed test).  If tail is 0,
    the alternative hypothesis is that the correlation is different
    than 0 (two tailed test).  If tail is -1, the alternative hypothesis
    is that the correlation is less than 0 (lower tailed test).
'''

docdict['n_permutations'] = '''
n_permutations : int, default: 1024
        Number of permutations.
'''

docdict['stat_fun_t'] = '''
stat_fun : callable() | None, default: None
    Function called to calculate the test statistic. Must accept 1D-array as
    input and return a 1D array. If ``None`` (the default), uses t statistic.
'''

docdict['stat_fun_F'] = '''
stat_fun : callable() | None, default: None
    Function called to calculate the test statistic. Must accept 1D-array as
    input and return a 1D array. If ``None`` (the default), uses t statistic
    or F statistic if more than two groups.
'''

docdict['stat_fun_corr'] = '''
stat_fun : callable() | None, default: None
    Function called to calculate the test statistic. Must accept 1D-array as
    input and return a 1D array. If ``None`` (the default), computes Pearson
    correlation.
'''

docdict['labels'] = '''
labels : array, shape (n_observations,)
    Condition label associated with each observation in ``X``.
'''

# overwrite MNE's documentation to avoid messing with bibtex for now
docdict['threshold'] = '''
threshold : float | dict | None, default: None
    The so-called "cluster forming threshold" in the form of a test statistic
    (note: this is not an alpha level / "p-value").
    If numeric, vertices with data values more extreme than ``threshold`` will
    be used to form clusters. If ``None``, threshold will be chosen
    automatically to correspond to a p-value of 0.05 for the given number of
    observations (only valid when using default statistic). If ``threshold`` is
    a :class:`dict` (with keys ``'start'`` and ``'step'``) then threshold-free
    cluster enhancement (TFCE) will be used (see TFCE example and [6]).
'''

docdict['alpha_spending_explanation'] = '''
Distributes Type I error over multiple, sequential analyses of the data (at
interim sample sizes specified in ``look_times`` never to exceed ``max_n``)
using a permutation-based adaptation of the alpha-spending procedure introduced
by Lan and DeMets [1]. This allows data collection to be terminated before
``max_n`` is reached if there is enough evidence to reject the null hypothesis
at an interim analysis, without inflating the false positive rate. This provides
a principled way to determine sample size and can result in substantial
efficiency gains over a fixed-sample design (i.e. can acheive the same
statistical power with a smaller expected sample size) [2, 3].
'''

docdict['references_clust'] = '''
.. [1] Gordon Lan, K. K., & DeMets, D. L. (1983).
    Discrete sequential boundaries for clinical trials.
    Biometrika, 70(3), 659-663.
.. [2] Lakens, D. (2014). Performing high-powered studies efficiently with
    sequential analyses. European Journal of Social Psychology, 44(7), 701-710.
.. [3] Lakens, D., Pahlke, F., & Wassmer, G. (2021).
    Group Sequential Designs: A Tutorial.
    https://doi.org/10.31234/osf.io/x4azm
.. [4] Maris, E., & Oostenveld, R. (2007). Nonparametric statistical testing of
    EEG-and MEG-data. Journal of neuroscience methods, 164(1), 177-190.
.. [5] Jona Sassenhagen and Dejan Draschkow. Cluster-based permutation tests of
    meg/eeg data do not establish significance of effect latency or location.
    Psychophysiology, 56(6):e13335, 2019. doi:10.1111/psyp.13335.
.. [6] Stephen M. Smith and Thomas E. Nichols. Threshold-free cluster
    enhancement: addressing problems of smoothing, threshold dependence and
    localisation in cluster inference.
    NeuroImage, 44(1):83–98, 2009. doi:10.1016/j.neuroimage.2008.03.061.
'''

docdict['references_maxtype'] = '''
.. [1] Gordon Lan, K. K., & DeMets, D. L. (1983).
    Discrete sequential boundaries for clinical trials.
    Biometrika, 70(3), 659-663.
.. [2] Lakens, D. (2014). Performing high‐powered studies efficiently with
    sequential analyses. European Journal of Social Psychology, 44(7), 701-710.
.. [3] Lakens, D., Pahlke, F., & Wassmer, G. (2021).
    Group Sequential Designs: A Tutorial.
    https://doi.org/10.31234/osf.io/x4azm
.. [4] Thomas E. Nichols and Andrew P. Holmes. Nonparametric permutation tests
    for functional neuroimaging: a primer with examples.
    Human Brain Mapping, 15(1):1–25, 2002. doi:10.1002/hbm.1058.
'''

docdict['returns_clust'] = '''
looks : dict
    Dictionary containing results of each look at the data, indexed by the
    values provided in ``look_times``. Each entry of the dictionary is a tuple
    that contains:

    ``obs`` : array, shape (p[, q][, r])
        Statistic observed for all variables.
    ``clusters`` : list
        List type defined by out_type above.
    ``cluster_pv`` : array
        P-value for each cluster.
    ``H0`` : array, shape (n_permutations,)
        Max cluster level stats observed under permutation.

ps : array, shape (n_looks,)
    The lowest p-value obtained at each look specied in ``look_times``. These
    can be compared to ``adj_alphas`` to determine on which looks, if any, one
    can reject the null hypothesis.
adj_alphas: array, shape (n_looks,)
    The adjusted significance thresholds for each look, chosen to control the
    false positive rate across multiple, sequential analyses. All p-values
    should be compared to the adjusted alpha for the look at which they were
    computed.
spending: array, shape (n_looks,)
    The value of the alpha spending function at each look.
'''

docdict['returns_maxtype'] = '''
looks : dict
    Dictionary containing results of each look at the data, indexed by the
    values provided in ``look_times``. Each entry of the dictionary is a tuple
    that contains:

    ``obs`` : array of shape (p[, q][, r])
        Test statistic observed for all variables.
    ``p_values`` : array of shape (p[, q][, r])
        P-values for all the tests (a.k.a. variables).
    ``H0`` : array of shape [n_permutations]
        Max test statistics obtained by permutations.

ps : array, shape (n_looks,)
    The lowest p-value obtained at each look specied in ``look_times``. These
    can be compared to ``adj_alphas`` to determine on which looks, if any, one
    can reject the null hypothesis.
adj_alphas: array, shape (n_looks,)
    The adjusted significance thresholds for each look, chosen to control the
    false positive rate across multiple, sequential analyses. All p-values
    should be compared to the adjusted alpha for the look at which they were
    computed.
spending: array, shape (n_looks,)
    The value of the alpha spending function at each look.
'''

docdict['alpha_spending_note'] = '''
The number of and timing of looks at the data need not be planned in advance
(other than ``n_max``), but it is important to include all looks that have
already occured in ``look_times`` each time you analyze the data to ensure that
valid adjusted significance thresholds are computed. In your final analysis,
``look_times`` should contain the ordered sample sizes at all looks at the data
that occured during the study.

When reporting results, you should minimally include the sample sizes at each
look, the minimum p-values at each look, the adjusted significance thresholds
for each look (to which the p-values are compared), and the value of the
alpha-spending function at each look. See [3] for further recommendations.
'''

##############################################################################
# Below is copied from pycrostates/utils/_docs.py
# Defines functions to fill in placeholders in docstrings.

docdict_indented: Dict[int, Dict[str, str]] = {}

def fill_doc(f: Callable) -> Callable:
    """Fill a docstring with docdict entries.
    Parameters
    ----------
    f : callable
        The function to fill the docstring of (modified in place).
    Returns
    -------
    f : callable
        The function, potentially with an updated __doc__.
    """
    docstring = f.__doc__
    if not docstring:
        return f

    lines = docstring.splitlines()
    indent_count = _indentcount_lines(lines)

    try:
        indented = docdict_indented[indent_count]
    except KeyError:
        indent = " " * indent_count
        docdict_indented[indent_count] = indented = dict()

        for name, docstr in docdict.items():
            lines = [
                indent + line if k != 0 else line
                for k, line in enumerate(docstr.strip().splitlines())
            ]
            indented[name] = "\n".join(lines)

    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split("\n")[0] if funcname is None else funcname
        raise RuntimeError(f"Error documenting {funcname}:\n{str(exp)}")

    return f


def _indentcount_lines(lines: List[str]) -> int:
    """Minimum indent for all lines in line list.
    >>> lines = [' one', '  two', '   three']
    >>> indentcount_lines(lines)
    1
    >>> lines = []
    >>> indentcount_lines(lines)
    0
    >>> lines = [' one']
    >>> indentcount_lines(lines)
    1
    >>> indentcount_lines(['    '])
    0
    """
    indent = sys.maxsize
    for k, line in enumerate(lines):
        if k == 0:
            continue
        line_stripped = line.lstrip()
        if line_stripped:
            indent = min(indent, len(line) - len(line_stripped))
    if indent == sys.maxsize:
        return 0
    return indent


def copy_doc(source: Callable) -> Callable:
    """Copy the docstring from another function (decorator).
    The docstring of the source function is prepepended to the docstring of the
    function wrapped by this decorator.
    This is useful when inheriting from a class and overloading a method. This
    decorator can be used to copy the docstring of the original method.
    Parameters
    ----------
    source : callable
        The function to copy the docstring from.
    Returns
    -------
    wrapper : callable
        The decorated function.
    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B(A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         ''' this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """

    def wrapper(func):
        if source.__doc__ is None or len(source.__doc__) == 0:
            raise RuntimeError(
                f"The docstring from {source.__name__} could not be copied "
                "because it was empty."
            )
        doc = source.__doc__
        if func.__doc__ is not None:
            doc += func.__doc__
        func.__doc__ = doc
        return func

    return wrapper



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

import numpy as np

def _correlation_stat_fun(X, y):
    '''
    computes correlation coefficients in vectorized manner for a substantial
    speedup vs. using scipy's implementation applied across axis

    Inputs
    -------
    X: an (n, num_tests) np.ndarray
    y: an (n, 1) np.ndarray

    Returns
    -------
    r: a (num_tests,) array of pearson's correlation coefficients
    '''
    Xm = np.mean(X,axis = 0)[np.newaxis, :]
    ym = np.mean(y)
    r_num = np.sum((X - Xm) * (y - ym), axis = 0)
    r_den = np.sqrt(np.sum((X - Xm)**2, axis = 0) * np.sum((y - ym)**2))
    r = r_num / r_den
    return r



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause


from ..spending_functions import SpendingFunction
from mne.parallel import parallel_func
from mne.utils import check_random_state
from inspect import isclass, isfunction
import numpy as np

from ..util.docs import fill_doc

def _boot_sample(x, n_resample, seed):
    '''
    performs two-level sampling of
        1. sample weights from direchlet distribution
        2. samples from x according to weights, with replacement

    Dirichlet distribution uses parameter alpha = 4, since this gives
    better approximations for the Bayesian bootstrap according
    to Tu and Zheng (1987).
    '''
    n_orig = x.shape[0]
    rng = check_random_state(seed)
    weights = rng.dirichlet(4 * np.ones(n_orig), 1)[0]
    idx = rng.choice(
        np.arange(n_orig), size = n_resample,
        replace = True, p = weights
    )
    x_boot = x[idx]
    return x_boot

def _boot_stat(x, statistic, n_resample, conditional = False, seed = 0, **kwargs):
    '''
    performs one bootstrap resampling and then computes statistic on new sample

    Equivalent to one draw of statistic from the posterior distribution
    '''
    if conditional:
        x_boot = _boot_sample(x, n_resample - x.shape[0], seed)
        _x = np.concatenate([x, x_boot])
    else:
        x_boot = _boot_sample(x, n_resample, seed)
        _x = x_boot
    kwargs['seed'] = seed
    return statistic(_x, **kwargs)

def _bootstrap(x, statistic, n_resample, n_simulations,
                conditional = False, seed = 0, n_jobs = None, **kwargs):
    '''
    performs full Bayesian bootstrap, optionally conditioned on current data
    '''
    # pick new random seeds to use for each permutation
    rng = check_random_state(seed)
    sim_seeds = rng.randint(1, np.iinfo(np.int32).max - 1, n_simulations)
    # perform `n_simulations` bootsrap resamplings
    parallel, p_func, _ = parallel_func(_boot_stat, n_jobs)
    out = parallel(
        p_func(x, statistic, n_resample, conditional, seed, **kwargs)
            for seed in sim_seeds
    )
    return np.stack(out)

_cumulative_power = lambda rejs: (np.cumsum(rejs, axis = 1) > 0).mean(0)

def _prob_rejection(rejs):
    assert(rejs.ndim == 2)
    first_rejs = np.zeros_like(rejs)
    for row in range(rejs.shape[0]):
        idxs = np.where(rejs[row, :])[0]
        if idxs.size > 0:
            idx = np.min(idxs)
            first_rejs[row, idx] = 1
    return first_rejs.mean(0)

def _expected_sample_size(rejs, look_times):
    idxs = [np.where(r)[0] for r in rejs]
    idxs = [np.min(i) if i.size > 0 else rejs.shape[1] - 1 for i in idxs]
    ns = [look_times[i] for i in idxs]
    return np.mean(ns)

@fill_doc
def bootstrap_predictive_power_1samp(X, test_func, look_times, n_max,
                alpha = .05, conditional = False, n_simulations = 1024,
                seed = None, n_jobs = None, **test_func_kwargs):
    '''Predictive power analysis via Bayesian bootstrap

    Computes the predictive power non-parametrically using the Bayesian
    bootstrap. Optionally, you can condition on the current data to get
    conditional power, which is useful for adaptive designs. Only valid for
    one-sample (or paired-sample) tests.

    Statistics computed from resamples using the Bayesian bootstrap, as opposed
    to the frequentist boostrap, can be interpreted as draws from the posterior
    distribution with an uninformative prior [1]. Thus, results here can be
    conveniently interpreted as the Bayesian predictive power. As recommended by
    [2] (not in English) and helpfully restated by [3] (in English), resampling
    weights are drawn from Dirichlet(alpha = 4) for a better approximation.

    This functionality is experimental. It is the best catch-all way to do a
    power analysis for permutation tests I can think of, and similar resampling
    approaches to estimating power have been used in the literature (e.g. by
    [4]); however, it should be noted that the neuroimaging literature has not
    converged upon a standardized approach to performing power analyses.
    The Bayesian bootstrap approach used here incorporates uncertainty about
    the effect size into the power estimate, which is handy since uncertainty
    about the true effect size is considerable following a small pilot study, or
    even a typical psychology/neuroimaging sample size, as pointed out by [5].

    Parameters
    -----------
    X : array, shape (n_observations, p[, q][, r])
        The data from which to resample. ``X`` should contain the observations
        for one group or paired differences. The first dimension of the array
        is the number of observations; remaining dimensions comprise the size of
        a single observation. See documentation for user-input ``test_func`` for
        more details.
    test_func : function
        The one-sample sequential test you want to run a power analysis for.
        Must accept ``look_times``, ``n_max``, and ``alpha`` arguments and
        return results, the middle two of which are the p-values for each look
        and the adjusted alphas, respectively. This could be any user-facing
        function from ``niseq`` that ends in ``_1samp``.
    %(look_times)s
    %(n_max)s
    %(alpha)s
    conditional : bool, default: ``False``
        If ``True``, performs a conditional power analysis; that is, computes
        the probability of a design rejecting the null hypothesis given that the
        data in ``X`` has already been collected and is included in the
        analysis, as in an adaptive design. If ``False`` (default), performs a
        prospective power analysis (e.g. if you're using pilot data or data from
        another study to inform sample size planning for a study that hasn't
        begun data collection).
    n_simulations: int, default: ``1024``
        Number of bootstrap resamples/simulations to perform.
    %(seed)s
    %(n_jobs)s
    **test_func_kwargs:
        You may input any arguments you'd like passed to ``test_func``.


    Returns
    --------
    res : dict
        A results dictionary with keys:

        ``'uncorr_instantaneous_power'`` : list of float
            The power of a fixed-sample statistical test performed at each look.
        ``'rejection_probability'`` : list of float
            The probability that a sequential test rejects the null hypothesis
            (for the first time) at each look time.
        ``'cumulative_power'`` : list of float
            The power of a sequential test to reject the null hypothesis by each
            look time. ``res['cumulative_power'][-1]`` is the power of the full
            sequential procedure.
        ``'uncorr_cumulative_power'`` : list of float
            Cumulative power if the rejection threshold at each look was not
            corrected using alpha-spending (as it should be).
        ``'n_expected'`` : float
            The expected sample size for the sequential procedure.
        ``'n_simulations'`` : int
            The number of bootstrap resamples used.
        ``'n_orig_data'`` : int
            The sample size of the original data ``X``, i.e. ``X.shape[0]``.
        ``'conditional'`` : bool
            Whether the power analysis that was run was conditional (``True``)
            or prospective (``False``).
        ``'test_func'`` : str
            Name of the sequential test function used.
        ``'test_func_kwargs'`` : dict
            A record of the arguments passed to the test function, including
            ``look_times`` and ``n_max``.

    Notes
    --------
    The significance level of the test used is specified in ``test_func``, and
    thus can be modified by passing an argument to ``test_func`` using
    ``**test_func_kwargs``.

    References
    ---------
    .. [1] Rubin, D. B. (1981). The bayesian bootstrap.
        The annals of statistics, 130-134.
    .. [2] Tu, D. & Zheng, Z. (1987). The Edgeworth expansion for the random
        weighting method. Chinese J. Appl. Probability and Statist., 3, 340-347.
    .. [3] Shao, J., & Tu, D. (2012). The jackknife and bootstrap.
        Springer Science & Business Media.
    .. [4] Ruzzoli, M., Torralba, M., Fernández, L. M., & Soto-Faraco, S. (2019).
        The relevance of alpha phase in human perception. Cortex, 120, 249-268.
    .. [5] Lakens, D., & Evers, E. R. (2014). Sailing from the seas of chaos
        into the corridor of stability: Practical recommendations to increase
        the informational value of studies.
        Perspectives on psychological science, 9(3), 278-292.





    '''
    def boot_stat(x, **kwargs): # thinly wrap test function
        _, ps, adj_alphas, _ = test_func(x, **kwargs)
        return np.stack([ps <= alpha, ps <= adj_alphas], axis = 1)
    test_func_kwargs['look_times'] = look_times
    test_func_kwargs['n_max'] = n_max
    test_func_kwargs['alpha'] = alpha
    rejections = _bootstrap(
        X, boot_stat,
        max(look_times), n_simulations,
        conditional, seed, n_jobs, **test_func_kwargs
    )

    # return simulation results with some metadata for posterity
    results = dict()
    results['uncorr_instantaneous_power'] = rejections[...,0].mean(0).tolist()
    results['rejection_probability'] = _prob_rejection(rejections[...,1]).tolist()
    results['cumulative_power'] = _cumulative_power(rejections[...,1]).tolist()
    results['uncorr_cumulative_power'] = _cumulative_power(rejections[...,0]).tolist()
    results['n_expected'] = _expected_sample_size(rejections[...,1], look_times)
    results['n_simulations'] = n_simulations
    results['n_orig_data'] = X.shape[0]
    results['conditional'] = False
    results['test_func'] = test_func.__name__
    test_func_kwargs = { # convert functions and classes to strings
        key: test_func_kwargs[key].__name__
            if (isfunction(test_func_kwargs[key]) or isclass(test_func_kwargs[key]))
            else test_func_kwargs[key]
            for key in test_func_kwargs
    }
    test_func_kwargs = { # convert SpendingFunction instances to strings
        key: test_func_kwargs[key].__class__.__name__
            if issubclass(type(test_func_kwargs[key]), SpendingFunction)
            else test_func_kwargs[key]
            for key in test_func_kwargs
    }
    results['test_func_kwargs'] = test_func_kwargs
    return results



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from ._permutation import generate_permutation_dist, find_thresholds
from ._clustering import _get_cluster_stats, _get_cluster_pvs
from ._permutation import _get_cluster_stats_correlation
import numpy as np

from .util.docs import fill_doc

@fill_doc
def sequential_cluster_test_1samp(X, look_times, n_max, alpha = .05, tail = 0,
                                    spending_func = None, **kwargs):
    '''A sequential one-sample cluster test.

    A sequential generalization of a one-sample cluster-based permutation test
    (as described by [4]) or of TFCE (as described by [6]).

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X_clust)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail)s
    %(spending_func)s
    %(threshold)s
    %(n_permutations)s
    %(stat_fun_clust_t)s
    %(adjacency_clust_1)s
    %(n_jobs)s
    %(seed)s
    %(max_step_clust)s
    %(exclude_clust)s
    %(t_power_clust)s
    %(out_type_clust)s
    %(check_disjoint_clust)s

    Returns
    ---------
    %(returns_clust)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_clust)s




    '''
    assert(isinstance(X, np.ndarray))
    obs, H0 = generate_permutation_dist(X, None, look_times, tail = tail, **kwargs)
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    obs_stats, ps = _get_cluster_pvs(obs, H0, tail = tail)
    return obs_stats, ps, adj_alpha, spending


@fill_doc
def sequential_cluster_test_indep(X, labels, look_times, n_max, alpha = .05,
                                tail = 0, spending_func = None, **kwargs):
    '''A sequential independent-sample cluster test.

    A sequential generalization of an independet-sample cluster-based
    permutation test (as described by [4]) or of TFCE (as described by [6]).

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X_clust)s
    %(labels)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail_clust)s
    %(spending_func)s
    %(threshold)s
    %(n_permutations)s
    %(tail)s
    %(stat_fun_clust_f)s
    %(adjacency_clust_1)s
    %(n_jobs)s
    %(seed)s
    %(max_step_clust)s
    %(exclude_clust)s
    %(f_power_clust)s
    %(out_type_clust)s
    %(check_disjoint_clust)s

    Returns
    ---------
    %(returns_clust)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_clust)s




    '''
    assert(isinstance(X, np.ndarray))
    assert(isinstance(X, np.ndarray))
    obs, H0 = generate_permutation_dist(
        X, labels, look_times,
        tail = tail,
        **kwargs
    )
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    tail = obs[look_times[0]][-2] # in case tail was overridden
    obs_stats, ps = _get_cluster_pvs(obs, H0, tail = tail)
    return obs_stats, ps, adj_alpha, spending


@fill_doc
def sequential_cluster_test_corr(X, y, look_times, n_max, alpha = .05, tail = 0,
                                    spending_func = None, **kwargs):
    '''A sequential cluster test for correlations.

    A sequential generalization of a cluster-based permutation test
    (as described by [4]) or of TFCE (as described by [6]) for testing a
    relationship between ``X`` and a continuous variable ``y``. Uses Pearson
    correlation by default (or its z-transform if using TFCE), but test
    statistic can be modified.

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X_clust)s
    %(y)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail)s
    %(spending_func)s
    %(threshold)s
    %(n_permutations)s
    %(tail_corr)s
    %(stat_fun_corr)s
    %(adjacency_clust_1)s
    %(n_jobs)s
    %(seed)s
    %(max_step_clust)s
    %(exclude_clust)s
    %(f_power_clust)s
    %(out_type_clust)s
    %(check_disjoint_clust)s

    Returns
    ---------
    %(returns_clust)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_clust)s




    '''
    assert(isinstance(X, np.ndarray) and isinstance(y, np.ndarray))
    obs, H0 = generate_permutation_dist(
        X, y, look_times,
        tail = tail,
        statistic = _get_cluster_stats_correlation,
        **kwargs
    )
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    obs_stats, ps = _get_cluster_pvs(obs, H0, tail = tail)
    return obs_stats, ps, adj_alpha, spending



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from ._permutation import generate_permutation_dist, find_thresholds
from mne.stats.parametric import f_oneway, ttest_1samp_no_p, ttest_ind_no_p
from mne.stats.cluster_level import _pval_from_histogram
from mne.utils import check_random_state, warn
from .util import _correlation_stat_fun
import numpy as np

from .util.docs import fill_doc

def _format_input(X):
    X = [x[:, np.newaxis] if x.ndim == 1 else x for x in X]
    sample_shape = X[0][0].shape
    X = [np.reshape(x, (x.shape[0], -1)) for x in X]
    return X, sample_shape


def _compute_max_stat(t_obs, tail):
    if tail == 0:
        max_stat = np.nanmax(np.abs(t_obs))
    elif tail == 1:
        max_stat = np.nanmax(t_obs)
    elif tail == -1:
        max_stat = np.nanmin(t_obs)
    return max_stat


def _get_max_stat_samples(X, labels = None,
                        statfun = None, tail = 0, **statfun_kwargs):
    '''
    Computes observed univariate stats using statfun for observed data.

    Inputs
    -------
    X: an (n_samples, <sample_shape>) np.ndarray
    labels: for independent samples test, an (n_samples,) np.ndarray containing
            condition labels, otherwise None for one-sample test
    statfun: a function to compute test statistics from X
    tail: sidedness of test... 0, 1, or -1

    Returns
    -------
    t_obs: a (<sample_shape>) np.ndarray containing test statistics
    t_max: the maximum test statistic





    '''
    assert(isinstance(X, np.ndarray))
    if statfun is None:
        raise Exception('must specify statfun!')
    if labels is None:
        X = [X]
    else:
        conds = np.unique(labels)
        X = [X[labels == cond] for cond in conds]
    X, sample_shape = _format_input(X)
    t_obs = statfun(*X, **statfun_kwargs)
    max_stat = _compute_max_stat(t_obs, tail)
    return t_obs.reshape(sample_shape), max_stat


def _get_max_stat_correlation(X, y, statfun, tail = 0, **statfun_kwargs):
    '''
    Inputs
    -------
    X: an (n_samples, <sample_shape>) np.ndarray
    y: an (n_samples,) np.ndarray
    statfun: a function to compute test statistics from X and y
    tail: sidedness of test... 0, 1, or -1

    Returns
    -------
    r_obs : a (<sample_shape>) np.ndarray containing test statistics
    r_max: the maximum test statistic





    '''
    X = [X, y]
    X, sample_shape = _format_input(X)
    r_obs = statfun(*X, **statfun_kwargs)
    max_stat = _compute_max_stat(r_obs, tail)
    return r_obs.reshape(sample_shape), max_stat


def _get_pvs_from_histogram(T_obs, H0, tail = 0):
    obs_shape = T_obs.shape
    T_obs = T_obs.flatten()
    if tail == 0:
        p_values = (H0 >= np.abs(T_obs[:, np.newaxis])).mean(-1)
    elif tail == 1:
        p_values = (H0 >= T_obs[:, np.newaxis]).mean(-1)
    elif tail == -1:
        p_values = (H0 <= T_obs[:, np.newaxis]).mean(-1)
    return p_values.reshape(obs_shape)


def _add_pvs_to_obs(obs_stats, H0, tail):
    '''
    formats observed data and associated p-values in similar format to
    mne.stats.permutations.permutation_t_test for each look time in obs_stats
    '''
    formatted_obs = {}
    min_ps = []
    for i, look_time in enumerate(obs_stats):
        obs = obs_stats[look_time][0]
        h0 = H0[:, i]
        p = _get_pvs_from_histogram(obs, h0, tail)
        # p should never be zero in a permutation test, so if p == 0, then
        p[p == 0] = np.nan # stat must have been compared to NaN or Inf
        formatted_obs[look_time] = (obs, p, h0)
        min_ps.append(np.nanmin(p))  
    return formatted_obs, np.array(min_ps)

@fill_doc
def sequential_permutation_t_test_1samp(X,
                                        look_times, n_max,
                                        alpha = .05, tail = 0,
                                        spending_func = None,
                                        **kwargs):
    '''One-sample sequential permutation test with max-type correction.

    This is a sequential generalization of
    ``mne.stats.permutations.permutation_t_test``.

    Uses max-type correction for multiple comparisons [4].

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail)s
    %(spending_func)s
    %(n_permutations)s
    %(n_jobs)s
    %(seed)s

    Returns
    ---------
    %(returns_maxtype)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_maxtype)s





    '''
    assert(isinstance(X, np.ndarray))
    obs, H0 = generate_permutation_dist(
        X, None,
        look_times,
        tail = tail,
        statistic = _get_max_stat_samples,
        statfun = ttest_1samp_no_p,
        **kwargs
        )
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    obs_stats, ps = _add_pvs_to_obs(obs, H0, tail)
    return obs_stats, ps, adj_alpha, spending


@fill_doc
def sequential_permutation_test_indep(X, labels,
                                        look_times, n_max,
                                        alpha = .05, tail = 0,
                                        spending_func = None,
                                        **kwargs):
    '''Independent-sample sequential permutation test with max-type correction.

    By default, this is a sequential generalization of an independent sample
    max-t procedure if two groups and max-F procedure if more groups.

    Uses max-type correction for multiple comparisons [4].

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X)s
    %(labels)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail)s
    %(spending_func)s
    %(n_permutations)s
    %(n_jobs)s
    %(seed)s

    Returns
    ---------
    %(returns_maxtype)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_maxtype)s





    '''
    assert(isinstance(X, np.ndarray))
    assert(isinstance(labels, np.ndarray))
    if np.unique(labels).size == 2:
        sf = ttest_ind_no_p
    else:
        sf = f_oneway
        if tail != 1:
            warn('Ignoring argument "tail", performing 1-tailed F-test')
            tail = 1

    obs, H0 = generate_permutation_dist(
        X, labels,
        look_times,
        tail = tail,
        statistic = _get_max_stat_samples,
        statfun = sf,
        **kwargs
        )
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    obs_stats, ps = _add_pvs_to_obs(obs, H0, tail)
    return obs_stats, ps, adj_alpha, spending


@fill_doc
def sequential_permutation_test_corr(X, y,
                                    look_times, n_max,
                                    alpha = .05, tail = 0,
                                    spending_func = None,
                                    **kwargs):
    '''A sequential permutation test for correlations with a max-type correction.

    Tests for a relationship between ``X`` and a continuous independent variable
    ``y``. By default, uses Pearson correlation by default, but the test
    statistic can be modified.

    Uses max-type correction for multiple comparisons [4].

    %(alpha_spending_explanation)s

    Parameters
    ----------
    %(X)s
    %(y)s
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail_corr)s
    %(spending_func)s
    %(n_permutations)s
    %(n_jobs)s
    %(seed)s

    Returns
    ---------
    %(returns_maxtype)s

    Notes
    ---------
    %(alpha_spending_note)s

    References
    ----------
    %(references_maxtype)s





    '''
    assert(isinstance(X, np.ndarray))
    assert(isinstance(y, np.ndarray))
    obs, H0 = generate_permutation_dist(
        X, y,
        look_times,
        tail = tail,
        statistic = _get_max_stat_correlation,
        statfun = _correlation_stat_fun,
        **kwargs
        )
    spending, adj_alpha = find_thresholds(
        H0, look_times, n_max,
        alpha, tail, spending_func
        )
    obs_stats, ps = _add_pvs_to_obs(obs, H0, tail)
    return obs_stats, ps, adj_alpha, spending



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from mne.stats.parametric import f_oneway, ttest_1samp_no_p
from mne.stats.cluster_level import (
    _permutation_cluster_test,
    _setup_adjacency,
    _cluster_indices_to_mask,
    _cluster_mask_to_indices,
    _get_partitions_from_adjacency,
    _validate_type,
    _find_clusters,
    _reshape_clusters,
    _pval_from_histogram
)
from mne.utils import _check_option, warn
from .util import _correlation_stat_fun
from collections import OrderedDict
from scipy.stats import beta
import numpy as np

def _check_fun(X, stat_fun, threshold, tail = 0, kind = 'within'):
    '''
    Same as mne.stats.cluster_level._check_fun but less verbose
    '''
    from scipy import stats
    if kind == 'within':
        ppf = stats.t.ppf
        if threshold is None:
            if stat_fun is not None and stat_fun is not ttest_1samp_no_p:
                warn('Automatic threshold is only valid for stat_fun=None '
                     '(or ttest_1samp_no_p), got %s' % (stat_fun,))
            p_thresh = 0.05 / (1 + (tail == 0))
            n_samples = len(X)
            threshold = -ppf(p_thresh, n_samples - 1)
            if np.sign(tail) < 0:
                threshold = -threshold
        stat_fun = ttest_1samp_no_p if stat_fun is None else stat_fun
    else:
        assert kind == 'between'
        ppf = stats.f.ppf
        if threshold is None:
            if stat_fun is not None and stat_fun is not f_oneway:
                warn('Automatic threshold is only valid for stat_fun=None '
                     '(or f_oneway), got %s' % (stat_fun,))
            elif tail != 1:
                warn('Ignoring argument "tail", performing 1-tailed F-test')
                tail = 1
            p_thresh = 0.05
            dfn = len(X) - 1
            dfd = np.sum([len(x) for x in X]) - len(X)
            threshold = ppf(1. - p_thresh, dfn, dfd)
        stat_fun = f_oneway if stat_fun is None else stat_fun
    return stat_fun, threshold, tail

def _get_cluster_stats(X, threshold = None, max_step = 1,
        tail = 0, stat_fun = None, adjacency = None,
        exclude = None, t_power = 1,
        out_type = 'mask', check_disjoint = False):
    '''
    An auxilliary function to compute cluster statistics from observed data or
    from a single permutation

    mostly copied from mne.stats.cluster_level._permutation_cluster_test, but
    it doesn't perform a full permutation test, just gets the observed stats
    '''
    n_samples = X[0].shape[0]
    n_times = X[0].shape[1]
    sample_shape = X[0].shape[1:]

    _check_option('out_type', out_type, ['mask', 'indices'])
    _check_option('tail', tail, [-1, 0, 1])
    if not isinstance(threshold, dict):
        threshold = float(threshold)
        if (tail < 0 and threshold > 0 or tail > 0 and threshold < 0 or
                tail == 0 and threshold < 0):
            raise ValueError('incompatible tail and threshold signs, got '
                             '%s and %s' % (tail, threshold))

    # flatten the last dimensions in case the data is high dimensional
    X = [np.reshape(x, (x.shape[0], -1)) for x in X]
    n_tests = X[0].shape[1]

    if adjacency is not None and adjacency is not False:
        adjacency = _setup_adjacency(adjacency, n_tests, n_times)

    if (exclude is not None) and not exclude.size == n_tests:
        raise ValueError('exclude must be the same shape as X[0]')

    t_obs = stat_fun(*X)
    _validate_type(t_obs, np.ndarray, 'return value of stat_fun')

    ## Compute mass-univariate test stat
    #--------------------------------------
    if t_obs.size != np.prod(sample_shape):
        raise ValueError('t_obs.shape %s provided by stat_fun %s is not '
                         'compatible with the sample shape %s'
                         % (t_obs.shape, stat_fun, sample_shape))
    if adjacency is None or adjacency is False:
        t_obs.shape = sample_shape

    if exclude is not None:
        include = np.logical_not(exclude)
    else:
        include = None

    ## compute cluster statistic from univariate test results
    # ------------------------------------------------------
    # determine if adjacency itself can be separated into disjoint sets
    if check_disjoint is True and (adjacency is not None and
                                   adjacency is not False):
        partitions = _get_partitions_from_adjacency(adjacency, n_times)
    else:
        partitions = None
    out = _find_clusters(t_obs, threshold, tail, adjacency,
                         max_step = max_step, include = include,
                         partitions = partitions, t_power = t_power,
                         show_info = False)
    clusters, cluster_stats = out

    # The stat should have the same shape as the samples
    t_obs.shape = sample_shape

    # For TFCE, return the "adjusted" statistic instead of raw scores
    if isinstance(threshold, dict):
        t_obs = cluster_stats.reshape(t_obs.shape) * np.sign(t_obs)
        cluster_stats = t_obs.reshape(cluster_stats.shape)

    # convert clusters to old format
    if adjacency is not None and adjacency is not False:
        # our algorithms output lists of indices by default
        if out_type == 'mask':
            clusters = _cluster_indices_to_mask(clusters, n_tests)
    else:
        # ndimage outputs slices or boolean masks by default
        if out_type == 'indices':
            clusters = _cluster_mask_to_indices(clusters, t_obs.shape)

    clusters = _reshape_clusters(clusters, sample_shape)
    if len(clusters) == 0: # handle case in which no clusters are found
        cluster_stats = np.array([0])
    if tail == 0:
        max_stat = np.max(np.abs(cluster_stats))
    elif tail == 1:
        max_stat = np.max(cluster_stats)
    elif tail == -1:
        max_stat = np.min(cluster_stats)
    return t_obs, clusters, cluster_stats, max_stat


def _get_cluster_stats_samples(X, labels = None, threshold = None, max_step = 1,
        tail = 0, stat_fun = None, adjacency = None,
        exclude = None, t_power = 1,
        out_type = 'mask', check_disjoint = False):
    '''
    Computes cluster stats when design is one-sample or independent samples
    '''
    ##  check inputs
    # -------------------------
    assert(isinstance(X, np.ndarray))
    if labels is None:
        X = [X]
        stat_fun, threshold, tail = _check_fun(X[0], stat_fun, threshold, tail)
    else:
        conds = np.unique(labels)
        X = [X[labels == cond] for cond in conds]
        stat_fun, threshold, tail = _check_fun(X, stat_fun, threshold, tail, 'between')

    # check dimensions for each group in X (a list at this stage).
    X = [x[:, np.newaxis] if x.ndim == 1 else x for x in X]
    sample_shape = X[0].shape[1:]
    for x in X:
        if x.shape[1:] != sample_shape:
            raise ValueError('All samples mush have the same size')

    t_obs, clusters, cluster_stats, max_stat = _get_cluster_stats(
            X, threshold, max_step,
            tail, stat_fun, adjacency,
            exclude , t_power,
            out_type, check_disjoint)
    return t_obs, clusters, cluster_stats, tail, max_stat


def _correlation_stat_fun_tfce(X, y):
    r = _correlation_stat_fun(X, y)
    z = np.arctanh(r) # Fisher z-transform
    return z

def _check_fun_correlation(X, stat_fun, threshold, tail):
    n = X.shape[0]
    if stat_fun is None:
        if isinstance(threshold, dict): # use z-transformed correlations
            stat_fun = _correlation_stat_fun_tfce
        else:
            stat_fun = _correlation_stat_fun
            if threshold is None:
                # default to r value where p < 0.05
                alpha = .05 / (1 + (tail == 0))
                dist = beta(n/2 - 1, n/2 - 1, loc = -1, scale = 2)
                if tail == 0:
                    threshold = -dist.ppf(alpha / 2)
                elif tail == 1:
                    threshold = -dist.ppf(alpha)
                elif tail == -1:
                    threshold = dist.ppf(alpha)
    else:
        assert(threshold is not None) # if stat_fun given, must define threshold
    return stat_fun, threshold

def _get_cluster_stats_correlation(X, y, threshold = None, max_step = 1,
        tail = 0, stat_fun = None, adjacency = None,
        exclude = None, t_power = 1,
        out_type = 'mask', check_disjoint = False):
    '''
    Computes cluster stats when y is a regression outcome variable
    '''
    assert(isinstance(X, np.ndarray) and isinstance(y, np.ndarray))
    assert(y.ndim == 1)
    stat_fun, threshold = _check_fun_correlation(X, stat_fun, threshold, tail)
    return _get_cluster_stats([X, y], threshold, max_step,
            tail, stat_fun, adjacency,
            exclude , t_power,
            out_type, check_disjoint)


def _get_cluster_pvs(obs_stats, H0, tail):
    '''
    computes cluster p-values at each look time by comparing observed clusters
    to permutation distribution

    Inputs
    --------
    obs_stats: (dict) of tuples of (t_obs, clusters, cluster_stats)
                indexed by look time
    H0: (np.ndarray) joint null distribution of cluster stats across look times
    tail: (int) sidedness of test to compute p-value for

    Returns
    --------
    stats: (dict of tuples) indexed by look time, this dict contains output
            matching that of MNE's cluster-based permutation test for that
            look time (i.e. t_obs, clusters, cluster_ps, H0)
    p_values: (np.ndarray) smallest p-value at each look time
    '''
    obs = OrderedDict()
    min_ps = [] # minimum at each look time
    for i, n in enumerate(obs_stats):
        t_obs, clusters, cluster_stats = obs_stats[n][:3]
        h0 = H0[:, i]
        if len(clusters) > 0:
            cluster_pv = _pval_from_histogram(cluster_stats, h0, tail)
            min_ps.append(np.min(cluster_pv))
        else:
            cluster_pv = np.array([])
            min_ps.append(1.) # no clusters
        obs[n] = (t_obs, clusters, cluster_pv, h0)
    return obs, np.array(min_ps)



from ..bootstrap import _bootstrap
import numpy as np

def test_bootstrap(mu = 10, std = 3, n = 100, n_boot = 1000, tol = .2):

    np.random.seed(0)

    def mean_stat(x, seed):
        return np.mean(x)
    def std_stat(x, seed):
        return np.std(x)

    x = np.random.normal(mu, std, n)
    mu_post = _bootstrap(x, mean_stat, n, n_boot)
    std_post = _bootstrap(x, std_stat, n, n_boot)

    assert(np.isclose(mu, mu_post.mean(), atol = tol))
    assert(np.isclose(std, std_post.mean(), atol = tol))

    return None



from ..docs import copy_doc, fill_doc

def test_fill_doc():
    """Test decorator to fill docstring."""

    @fill_doc
    def foo(n_clusters, verbose):
        """My doc.
        Parameters
        ----------
        %(n_max)s
        """
        pass

    assert "n_max : int" in foo.__doc__
    

def test_copy_doc():
    """Test decorator to copy docstring."""

    def foo(x, y):
        """
        My doc.
        """
        pass

    @copy_doc(foo)
    def foo2(x, y):
        pass

    assert "My doc." in foo2.__doc__



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from ._clustering import _get_cluster_stats_samples, _get_cluster_stats_correlation
from .spending_functions import SpendingFunction, LinearSpendingFunction
from mne.utils import check_random_state
from mne.parallel import parallel_func
from collections import OrderedDict
import numpy as np

from .util.docs import fill_doc

def _get_stat_at_look_times(X, labels, look_times, statistic, statistic_kwargs):
    '''
    computes the test statistic at each look time
    '''
    obs = OrderedDict()

    for i, n in enumerate(look_times):

        _X = X[:n]

        if labels is None: # one-sample test
            obs[n] = statistic(_X, **statistic_kwargs)
        else: # independent-samples test
            _labels = labels[:n]
            obs[n] = statistic(_X, _labels, **statistic_kwargs)

    return obs


def _get_stat_perm(X, labels, look_times, seed, statistic, statistic_kwargs):
    '''
    gets the test statistics, at each look time, for a single permutation
    '''
    rng = check_random_state(seed)

    if labels is None: # one-sample, permute by sign flips
        flips = rng.choice([-1, 1], size = X.shape[0])
        while X.ndim > flips.ndim:
            flips = flips[:, np.newaxis]
        _X = X * flips
        _labels = None
    else: # independent-sample, permute by shuffling labels:
        # but only shuffle within each period between look times,
        # so p-values at each look time will be approximately preserved
        # across multiple, sequential uses of the test
        idxs = np.arange(X.shape[0])
        _lt = [0] + list(look_times)
        starts = [_lt[i] for i in range(len(look_times))]
        ends = [_lt[i + 1] for i in range(len(look_times))]
        for s, e in zip(starts, ends):
            rng.shuffle(idxs[s:e])
        _X = X
        _labels = labels[idxs]
    stats = _get_stat_at_look_times(
        _X, _labels,
        look_times,
        statistic, statistic_kwargs
    )
    H0 = [stats[n] for n in look_times]
    if isinstance(H0[0], tuple):
        H0 = [h0[-1] for h0 in H0]
    H0 = np.array(H0)
    assert(H0.ndim == 1)
    return H0

@fill_doc
def generate_permutation_dist(X, labels,
                            look_times,
                            n_permutations = 1024,
                            seed = None,
                            n_jobs = None,
                            statistic = _get_cluster_stats_samples,
                            **statistic_kwargs):
    '''
    This function computes the test statistic and its permutation distribution
    at each look time. It isn't meant for users to access directly for ordinary
    use, though it can be used in combination with ``find_thresholds`` to
    construct new sequential tests if you're confident you know what you're
    doing. You'll want to read the source code carefully to make sure your
    ``statistic`` function is compatible.

    Arguments
    -----------
    %(X)s
    labels : array of shape (n_observations,) | None
        Either condition labels for each observation in ``X``, a continuous
        dependent variable to correlate with ``X``, or None. In the latter case,
        a one-sample (sign flip) permutation scheme will be used, otherwise an
        independent sample (label shuffle) permutation scheme is used.
    %(n_permutations)s
    %(seed)s
    %(n_jobs)s
    statistic : callable(), default: _get_cluster_stats_samples
        The test statistic to compute on the data, e.g. a cluster statistic or
        a max-t statistic. The last value ``statistic`` returns must be the
        omnibus test statistic (e.g. the max-t or the cluster size), though you
        can return whatever other stuff you want which will be passed through
        the ``obs`` dictionary.
    **statistic_kwargs :
        You may pass arbitrary arguments to the ``statistic`` function.

    Returns
    ----------
    obs : dict
        The output of `statistic` indexed by look time in ``look_times``.
    H0 : array of shape (n_permutations, n_looks)
        The joint permutation null distribution of the test statistic across
        look times.




    '''
    # pick new random seeds to use for each permutation
    rng = check_random_state(seed)
    perm_seeds = rng.randint(1, np.iinfo(np.int32).max - 1, n_permutations)

    look_times = sorted(look_times)
    last_look = look_times[-1]
    assert(last_look <= X.shape[0])

    obs = _get_stat_at_look_times(
        X, labels,
        look_times,
        statistic, statistic_kwargs
    )
    obs_stats = [obs[n] for n in look_times]
    if isinstance(obs_stats[0], tuple): # test stat should be last
        obs_stats = [s[-1] for s in obs_stats]
    obs_stats = np.array(obs_stats)

    parallel, p_func, n_jobs = parallel_func(_get_stat_perm, n_jobs)
    out = parallel(
        p_func(
            X, labels,
            look_times,
            ps,
            statistic, statistic_kwargs
        ) for ps in perm_seeds
    )
    H0 = np.stack(out)
    H0 = np.concatenate([obs_stats[np.newaxis, :], H0])
    return obs, H0


def _quantile(x, q, method):
    '''
    deal with older numpy versions that only use deprecated `interpolation` arg
    '''
    try:
        return np.quantile(x, q, method = method)
    except:
        return np.quantile(x, q, interpolation = method)

def _throw_spending_func_warning():
    '''
    raised when user spending function doesn't match specified alpha and n_max
    '''
    import warnings
    warnings.warn(
        '''
        User input spending_func doesn't match `alpha` and `n_max` arguments!
        `alpha` and `n_max` have been overridden by those of `spending_func`.
        '''
    )

@fill_doc
def find_thresholds(
    H0, look_times, max_n,
    alpha = 0.05, tail = 0,
    spending_func = None
    ):
    '''
    Given a permutation null distribution for a corresponding sequence of look
    times and an alpha spending function, computes the adjusted significance
    thresholds requires to control the false positive rate across all looks.

    This isn't meant to be accessed directly by users, but it can be used
    together with ``generate_permutation_dist`` to create new sequential tests
    if you're confident you know what you're doing.


    Arguments
    ---------
    H0 : array of shape (n_permutations, n_looks)
        The joint permutation null distribution of the test statistic across
        look times.
    %(look_times)s
    %(n_max)s
    %(alpha)s
    %(tail)s
    %(spending_func)s

    Returns
    --------
    spending : array of shape (n_looks,)
        The value of the alpha spending function at each sample size
        in ``look_times``.
    adj_alphas : array of shape (n_looks,)
        The adjusted significance threshold against which to compare p-values
        at each sample size in ``look_times``.




    '''
    # check spending function
    if spending_func is None:
        spending_func = LinearSpendingFunction(alpha, max_n)
    else:
        assert(isinstance(spending_func, SpendingFunction))
        try:
            assert(spending_func.alpha == alpha)
            assert(spending_func.max_n == max_n)
        except:
            alpha = spending_func.alpha
            max_n = spending_func.max_n
            _throw_spending_func_warning()

    look_times = sorted(look_times)
    last_look = look_times[-1]
    assert(last_look <= max_n)


    _H0 = np.copy(H0)
    spending_hist = []
    adjusted_alphas = []
    assert(isinstance(spending_func, SpendingFunction))
    for i, n in enumerate(look_times):
        budget = spending_func(n)
        if tail == 1:
            thres = _quantile(_H0[:, i], 1 - budget, 'higher')
            adjusted_alpha = np.mean(H0[:, i] > thres)
            # If a permutation exceeds threshold at one look time...
            perm_false_positives = _H0[:, i] >= thres
            # ... we pass it forward so counted against future looks
            _H0[perm_false_positives, :] = np.inf
        elif tail == -1:
            thres = _quantile(_H0[:, i], budget, 'lower')
            adjusted_alpha = np.mean(H0[:, i] < thres)
            perm_false_positives = _H0[:, i] <= thres
            _H0[perm_false_positives, :] = -np.inf
        else:
            thres = _quantile(np.abs(_H0[:, i]),1 - budget, 'higher')
            adjusted_alpha = np.mean(np.abs(H0[:, i]) > thres)
            perm_false_positives = np.abs(_H0[:, i]) >= thres
            _H0[perm_false_positives, :] = np.inf
        spending_hist.append(budget)
        adjusted_alphas.append(adjusted_alpha)
    # the first adjusted alpha should always be equal to the spending function
    # at the first look, but the empircal threshold computed here may differ
    # numerically, especially e.g. in cases where the test statistic is discrete
    # like for NBS, so we correct it here:
    adjusted_alphas[0] = spending_hist[0]
    return np.array(spending_hist), np.array(adjusted_alphas)



from mne.utils import check_random_state
from mne.parallel import parallel_func
from ..max_test import (
    sequential_permutation_t_test_1samp,
    sequential_permutation_test_indep,
    sequential_permutation_test_corr
)
import numpy as np

N_SIMULATIONS = 500

def one_simulation(seed, tail = 0, sim_type = '1samp',
                   look_times = np.linspace(100, 300, 3).astype(int)):

    # generate null data
    rng = check_random_state(seed)
    x = rng.normal(loc = 0, size = look_times[-1])

    # run sequential test
    if sim_type == 'indep':
        conds = rng.choice([0, 1], look_times[-1])
        _, p, adj_alpha, _ = sequential_permutation_test_indep(
            x, conds, look_times, n_max = look_times[-1],
            tail = tail,
            seed = seed
        )
    elif sim_type == '1samp':
        _, p, adj_alpha, _ = sequential_permutation_t_test_1samp(
            x, look_times, n_max = look_times[-1],
            tail = tail,
            seed = seed
        )
    elif sim_type == 'corr':
        y = rng.rand(look_times[-1])
        _, p, adj_alpha, _ = sequential_permutation_test_corr(
            x, y, look_times, n_max = look_times[-1],
            tail = tail,
            seed = seed
        )

    # reject if p-val crosses sig threshold at any look time
    return np.array([np.any(p < .05), np.any(p < adj_alpha)])

def fpr_by_simulation(n_simulations, tail, sim_type, n_jobs = -1):
    parallel, p_func, _ = parallel_func(one_simulation, n_jobs)
    out = parallel(p_func(seed, tail, sim_type) for seed in range(n_simulations))
    rejections = np.stack(out)
    fpr = rejections.mean(0)
    return fpr[1] # false positive rate with sequential correction



def test_1samp_fpr(n_simulations = N_SIMULATIONS, alpha = .05, slack = .01):
    np.random.seed(0)
    for tail in [0, 1, -1]:
        fpr = fpr_by_simulation(n_simulations, tail, sim_type = '1samp')
        assert(fpr <= alpha + slack)
    return None

def test_indep_fpr(n_simulations = N_SIMULATIONS, alpha = .05, slack = .01):
    np.random.seed(0)
    for tail in [0, 1, -1]:
        fpr = fpr_by_simulation(N_SIMULATIONS, tail, sim_type = 'indep')
        assert(fpr <= alpha + slack)
    return None

def test_corr_fpr(n_simulations = N_SIMULATIONS, alpha = .05, slack = .01):
    np.random.seed(0)
    for tail in [0, 1, -1]:
        fpr = fpr_by_simulation(n_simulations, tail, sim_type = 'corr')
        assert(fpr <= alpha + slack)
    return None



from setuptools import setup, find_packages
import sys 

setup(
    name = 'niseq',
    version = '0.0.1',
    description = 'Group sequential tests for neuroimaging',
    url = 'https://github.com/john-veillette/mne-ari',
    author = 'John Veillette',
    author_email = 'johnv@uchicago.edu',
    license = 'BSD-3-Clause',
    packages = find_packages(),
    install_requires = ['mne>=1.0'],
    classifiers = [
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved',
            'Programming Language :: Python',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Operating System :: MacOS',
            'Programming Language :: Python :: 3',
    ]
)



# -*- coding: utf-8 -*-

# Authors: John Veillette <johnv@uchicago.edu>
#
# License: BSD-3-Clause

from abc import ABC, abstractmethod
import numpy as np

from .util.docs import copy_doc

class SpendingFunction(ABC):
    '''An abstract base class for spending functions
    '''
    def __init__(self, alpha, max_n):
        '''
        Parameters
        -----------
        alpha : float
            Desired false-positive rate after all sequential tests.
        max_n : int
            The sample size at which data collection will be terminated,
            regardless of whether the null hypothesis has been rejected.
        '''
        # save specifications
        self.alpha = alpha
        self.max_n = max_n
        # and test that spending function is valid, a.k.a. it . . .
        assert(np.isclose(self.__call__(0), 0)) # starts at zero
        assert(np.isclose(self.__call__(max_n), alpha)) # ends at alpha
        a = [self.__call__(n) for n in range(max_n)]
        assert(np.all(np.diff(np.array(a)) >= 0)) # and monotonically increases

    @abstractmethod
    def __call__(self, n):
        '''
        Parameters
        ----------
        n : int
            An interim sample size

        Returns
        ----------
        spending : float
            The value of the spending function at ``n``
        '''
        pass

class LinearSpendingFunction(SpendingFunction):
    '''The linear spending function.
    This is the simplest possible spending function, which
    distributes Type I error rate allowance evenly over time.
    '''
    @copy_doc(SpendingFunction.__call__)
    def __call__(self, n):
        return n/self.max_n * self.alpha

class PocockSpendingFunction(SpendingFunction):
    '''The Pocock spending function.
    A very common spending function that spends your alpha budget somewhat
    liberally (compared to e.g. the O'Brien-Fleming function) at the
    beginning of study; consequently, you have more power early on in
    exchange for a sharper penalty as you approach the maximum sample size.
    '''
    @copy_doc(SpendingFunction.__call__)
    def __call__(self, n):
        return self.alpha * np.log(1 + (np.e - 1)*(n/self.max_n))

class OBrienFlemingSpendingFunction(SpendingFunction):
    '''The O'Brien Fleming spending function.
    A common choice for clinical trials or other confirmatory research, this
    spending function is conservative for early analyses, saving more power
    for later in the study.
    '''
    @copy_doc(SpendingFunction.__call__)
    def __call__(self, n):
        from scipy.stats import norm
        r = n / self.max_n # information fraction
        if n != 0:
            return 2 * (1 - norm.cdf(norm.ppf(1 - self.alpha / 2) / np.sqrt(r)))
        else:
            return 0. # avoid divide by zero

class PiecewiseSpendingFunction(SpendingFunction):
    '''A piecewise spending function for adjusting the maximum sample size.
    A piecewise spending function to be used when adjusting your maximum sample
    size. i.e., the old spending function ``old_spending_func`` is used up until
    ``break_n``, the intermediate sample size at which you decided to change the
    max sample size. After that, a linear function is used that goes from
    ``old_spending_func(break_n)`` to (``new_max_n``, ``alpha``).

    This is useful if, for instance, (1) you accidentally collect more data than
    your original ``max_n``, requiring you to adjust your spending function, or
    (2) if a conditional power analysis encourages you to change your sample
    size to acheive a desired Type II error rate. Also (3) if you can no longer
    collect your original ``max_n`` for practical reasons.

    If ``max_n`` is adjusted multiple times, you can create piecewise spending
    functions recursively.
    '''

    def __init__(self, old_spending_func, break_n, new_max_n):
        '''
        Parameters
        -----------
        old_spending_func : instance of SpendingFunction
            an initialized instance of a SpendingFunction subclass with
            ``old_spending_func.max_n`` greater than ``break_n``.
        break_n : int
            The interim sample size at which researcher decides to adjust their
            maximum sample size.
        new_max_n : int
            The new maximum sample size.
        '''
        self.old_spending_func = old_spending_func
        self.break_n = break_n
        super().__init__(old_spending_func.alpha, new_max_n)

    @copy_doc(SpendingFunction.__call__)
    def __call__(self, n):
        if n <= self.break_n:
            return self.old_spending_func(n)
        else: # linear interpolation between spending at break point and max
            a = self.old_spending_func(self.break_n) # spending at break point
            m = (self.alpha - a)/(self.max_n - self.break_n)
            return m * (n - self.break_n) + a
